//SLAMM SOURCE CODE Copyright (c) 2009-2012 Warren Pinnacle Consulting, Inc.
//Code Use and Redistribution is Subject to Licensing, SEE SLAMM_License.txt
//

unit GLOBAL;

INTERFACE

USES   SysUtils, Windows, Messages, Classes, Graphics, Math, Dialogs;

Const  VersionNum: Double = 6.50;
       VersStr  = '6.2.0 beta';
       BuildStr = '6.2.0001';

Const  Feet_Per_Meter = 3.28084;

Const USE_DATAELEV = TRUE; {Use an array that displays elevations above 8 meters}

CONST  MTL =  0.0;         {use mean tidal level as datum}
       ELEV_CUTOFF = 8.0;  {elev not to model when addtl optimize is turned on, in meters above MTL}

Const Tiny = 1e-10;

VAR DRIVE_EXTERNALLY : Boolean = FALSE;

TYPE
  TLine = packed record p1,p2: Tpoint end;                                
  WaveDirection = (Easterly,Westerly,Southerly,Northerly);

  Category =
   ( DevDryland,  	    { 1 U	                         Upland}
     UndDryland,	      { 2 U	                         Upland  - default; these two will have to be distinguished using census or land use data (protection)}
     Swamp,	            { 3 PFO, PFO1, PFO3-5, PSS           Palustrine forested broad-leaved deciduoua}
     CypressSwamp,	    { 4 PFO2	                           Ditto, needle-leaved deciduous}
     InlandFreshMarsh,  { 5 L2EM,PEM[1&2]["A"-"I"],R2EM      Lacustrine, Palustrine, and Riverine emergent}
     TidalFreshMarsh,	  { 6 R1EM,PEM["K"-"U"]                Riverine tidal emergent}
     ScrubShrub,        { 7 E2SS1, E2FO                      Estuarine intertidal scrub-shrub broad-leaved deciduous}
     RegFloodMarsh,	    { 8 E2EM [no "P"]                    Estuarine intertidal emergent [won't distinguish high and low marsh]}
     Mangrove,	        { 9 E2FO3, E2SS3                     Estuarine intertidal forested and scrub-shrub broad-leaved evergreeen}
     EstuarineBeach,	  {10 E2US2 or E2BB (PUS"K")           Estuarine intertidal unconsolidated shore sand or beach-bar}
     TidalFlat,	        {11 E2US3or4, E2FL, M2AB             Estuarine intertidal unconsolidated shore mud/organic or flat, Intertidal Aquatic Bed here for now}
     OceanBeach,	      {12 M2US2, M2BB/UB/USN               Marine intertidal unconsolidated shore sand}
     OceanFlat,	        {13 M2US3or4                         Marine intertidal unconsolidated shore mud or organic}
     RockyIntertidal,	  {14 M2RS, E2RS, L2RS                 Estuarine & Marine intertidal rocky shore}
     InlandOpenWater,   {15 R3-UB R2-5OW,L1-2OW,POW,PUB,R2UB
                            (L1-2,UB,AB), PAB, R2AB          Riverine, Lacustrine, and Palustrine open water}
     RiverineTidal,     {16 R1OW	                           Riverine tidal open water}
     EstuarineWater,    {17 E1, (PUB"K" no "h")              Estuarine subtidal}
     TidalCreek,	      {18 E2SB, E2UBN,                     Estuarine intertidal stream bed}
     OpenOcean,         {19 M1                               Marine subtidal  [aquatic beds and reefs can be added later]}
     IrregFloodMarsh,   {20 E2EM[1-5]P                       "P" Irregularly Flooded Estuarine Intertidal Emergent}
     NotUsed,           {21 [No Codes]                       Was "Tall Spartina" now obsolete }
     InlandShore,       {22 L2UD,PUS, R[1..4]US/RS           shoreline not pre-processed using Tidal Range Elevations}
     TidalSwamp,        {23 PSS,PFO"K"-"V" /EM1"K"-"V"       Tidally influenced Swamp. }

     Blank,             {24}
     VegTidalFlat,      {25  E2AB3, or TNC specific Data}
     BackShore);        {26  Some salt influence, TNC specific Data }

   ElevCategory =
    ( LessThanMLLW,
      MLLWtoMLW,
      MLWtoMTL,
      MTLtoMHW,
      MHWtoMHWS,
      GreaterThanMHWS);

Const  FirstCategory = DevDryLand;
       LastCategory  = BackShore;


TYPE

  ElevUnit = (HalfTide, SaltBound, Meters);

  ClassElev = packed Record
      MinUnit : ElevUnit;
      MinElev : Double;
      MaxUnit : ElevUnit;
      MaxElev : Double;
    End; {Record}


Type TClassElevArray = Array[FirstCategory..LastCategory] of ClassElev;

Procedure LoadElevRangesFromText(Var TCA: TClassElevArray);
Procedure SaveElevRangesToText(TCA: TClassElevArray);

Type
 TPointArray  = Array of TPoint;
// PPointArray = ^TPointArray;

 TPolygon  = Class
               TPoints: TPointArray;
               NumPts: Integer;
               svMinRow,svMaxRow,svMinCol,svMaxCol: Integer;  //no save
               Function InPoly(Row,Col: Integer): Boolean;
               Function MinRow: Integer;
               Function MaxRow: Integer;
               Function MinCol: Integer;
               Function MaxCol: Integer;
               Constructor Create;
               Constructor Load(ReadVersionNum: Double; TS: TStream; ReadHeader: Boolean);
               Procedure Store;
               Destructor Destroy; override;
             End;

 TRectangle = Packed Record
    X1,Y1,X2,Y2   : Integer;  {rectangular location within primary map}
  End;


 POutputSite2 = ^TOutputSite2;
 TOutputSite2 = Packed Record
    UsePolygon: Boolean;
    Rec: TRectangle;
    Poly: TPolygon;
    Description   : String;
  End;



Const N_ACCR_MODELS = 4;
      AccrNames : Array[0..N_ACCR_MODELS-1] of String = ('Reg Flood','Irreg Flood','T.Flat','Tidal Fresh');  //1-23-11

Type
 TUseAccrVar =  Array[0..N_ACCR_MODELS-1] of Boolean;
 TAccrVar =  Array[0..N_ACCR_MODELS-1] of Double;
 TAccrNotes = Array[0..N_ACCR_MODELS-1] of String;
  {0 = Reg Flood, 1 = Irreg Flood, 2 = Tidal Flat, 3 = veg Tidal Flat}


 TSubSite = Class
    Poly: TPolygon;
    Description   : String;
    NWI_Photo_Date: Integer;
    DEM_date  : Integer;
    Direction_OffShore  : WaveDirection; { S, E, W, N }
    Historic_trend,          { (mm/yr) assumed to include subsidence   }
    NAVD88MTL_correction,    { m, MTL-NAVD88 in meters                 }
    GTideRange,              { m, Great Diurnal Tide Range, GT         }
    SaltElev  : double;      { m above MTL, was Mean High Water Spring }
    MarshErosion,
    SwampErosion,
    TFlatErosion  : Double;  {horizontal erosion meters / year}
    FixedRegFloodAccr,          {accretion rates, mm / year}
    FixedIrregFloodAccr,        {accretion rates, mm / year}
    FixedTideFreshAccr: Double; {accretion rates, mm / year}
    InlandFreshAccr,            {New 5/20/2011 }
    MangroveAccr,               {New 5/20/2011 }
    TSwampAccr, SwampAccr : Double;   {New 5/20/2011 }

    Fixed_TF_Beach_Sed: Double;  {Sedimentation for non-wetlands, m /year}
    Freq_BigStorm: Double;
    Use_Preprocessor: Boolean;  {whether to use elev. pre-processor, i.e. TRUE using poor vertical resolution data FALSE if using LiDAR}

    OW_Max_Width_Overwash: Double;  {adds flexibility to SLAMM overwash model}
    OW_Beach_to_Ocean: Double;
    OW_Dryland_to_Beach: Double;
    OW_Est_to_Beach: Double;
    OW_Marsh_Pct_Loss: Double;
    OW_Mang_Pct_Loss: Double;

    {Accretion Submodel Parameters}   {0 = Reg Flood, 1 = Irreg Flood, 2 = Tidal Flat, 3 = veg Tidal Flat}
    UseAccrModel: TUseAccrVar;  {boolean}
    MaxAccr : TAccrVar;  {mm/year}
    MinAccr : TAccrVar; {mm/year}
    AccrA   : TAccrVar; {unitless}
    AccrB   : TAccrVar; {unitless}
    AccrC   : TAccrVar; {unitless}
    DistEffectMax: TAccrVar; {meters}
    Dmin: TAccrVar; {unitless scaling factor, set to 1.0 to ignore distance factor}
    SalinityTMax : TAccrVar; {ppt}
    STMaxZone    : TAccrVar; {ppt}
    SNonTMax : TAccrVar; {unitless multiplier, set to one to ignore salinity}
    AccrNotes: TAccrNotes;

    {Calculations below here}
    MHHW,
    MLLW,
    Norm      : Double;  { Global SLR since model start, cm }
    SLRise,              { Rise of SL in m per time-step    }
    OldSL,
    NewSL    : Double;
    DeltaT   : Integer;
    LastOverwash : Integer;
    OverWashes_This_dT: Integer;

    Constructor Create;
    Constructor CreateWith(TSS: TSubSite);
    Destructor Destroy;  override;

  End;


  TSite = Class
    Rows, Cols    : Integer;  { Number of rows & columns }
    LLXCorner,                { SW corner of site        }
    LLYCorner     : Double;
    Scale         : double;   { width & height of square cell cells }

    GlobalSite    : TSubSite; { site data used if no "subsites" or no "subsites" meet boundary }
    NSubSites,NOutputSites  : Integer;
    MaxROS        : Integer;  { ROS = raster output sites }
    SubSites      : Array of TSubSite;
    OutputSites  : Array of TOutputSite2;
    ROSBounds     : Array of TRectangle;  {read from rasters}
    NoElevData : double;
    Constructor Create;
    Constructor Load(ReadVersionNum: Double; TS: TStream);
    Procedure Store(Var TS: TStream);
    Procedure LoadSubSite(ReadVersionNum: Double; TS: TStream; Var SS: TSubSite);
    Procedure StoreSubSite(Var TS: TStream; Var SS: TSubSite);
    Procedure AddSubSite;
    Procedure AddOutputSite;
    Procedure DelSubSite(Index: Integer);    {indexed to zero}
    Procedure DelOutputSite(Index: Integer); {indexed to zero}
    Function GetSubSite(X {Col},Y {Row}:Integer): TSubSite;  {Subsite with highest number gets dominance} overload;
    Function GetSubSite(X,Y:Integer; PC: pointer): TSubSite;  {Subsite with highest number gets dominance} overload;
    Function GetSubSiteNum(X,Y:Integer): Integer; overload;
    Function GetSubSiteNum(X,Y:Integer; PC: pointer): Integer; overload;
    Function InOutSite(X {Col},Y {Row}:Integer; SiteNum: Integer): Boolean;
    Function T0 : Integer; {Latest NWI_Photo_Date}
    Function TideOc(X,Y: Integer): Double;
    Procedure InitElevVars;
    Destructor Destroy; override;
  End;

Type StatRecord = Packed Record
    Min, Max, Sum, SumX2, Sum_e2, Mean, Stdev, p05, p95 : Array[Category] of Double;
End; {Record}

Const HistWidth = 100 {100};  {used for calculating salinity histograms, n.b. affects STORE/LOAD}

Type TElevStats = Packed Record
    Derived: TDateTime;
    N: Array[Category] of Int64;
    Stats: Array[1..8] of StatRecord;
    Notes: Array[Category] of Integer;  //Not Used
    pLowerMin: Array[Category] of Double;
    Histogram : Array[1..8,Category,0..(HistWidth-1)] of Integer;   //1/5:HTU/m-MTL, 2/6:HTU/m-MLLW, 3/7:HTU/m-MHHW, 4/8:HTU/m-NAVD88
  End; {Record}

 TSalinityRule = class
    FromCat, ToCat: Category;
    SalinityLevel: Double;
    GreaterThan: Boolean;
    SalinityTide: Integer; { 1 = MLLW, 2 = MTL, 3 = MHHW, 4 = Monthly High }
    Descript: String;
    Constructor Load(ReadVersionNum: Double; TS: TStream);
    Procedure Store(Var TS: TStream);
    Constructor Create;
 End;

 TSalinityRules = class
   NRules: Integer;
   Rules : Array of TSalinityRule;
   Function FindRule(Cat: Category; RuleNum: Integer): TSalinityRule;
   Function CountRules(Cat: Category): Integer;
   Constructor Create;
   Constructor Load(ReadVersionNum: Double; TS: TStream);
   Procedure Store(Var TS: TStream);
 End;

Type GridColorType = ARRAY[Category] OF TColor;
     GridColor2Type = Array[Category] of TRGBTriple;
Type ElevGridColorType = ARRAY[ElevCategory] OF TColor;

Const
  TitleCat   : ARRAY[Category] OF STRING
    =('Developed Dry Land','Undeveloped Dry Land','Swamp','Cypress Swamp','Inland-Fresh Marsh',
      'Tidal-Fresh Marsh','Trans. Salt Marsh','Regularly-Flooded Marsh','Mangrove','Estuarine Beach',
      'Tidal Flat','Ocean Beach','Ocean Flat','Rocky Intertidal','Inland Open Water',
      'Riverine Tidal','Estuarine Open Water','Tidal Creek','Open Ocean','Irreg.-Flooded Marsh',
      'Not Used','Inland Shore','Tidal Swamp','Blank','Vegetated Tidal Flat','Backshore');

  DefaultGridColors: GridColorType
   = ($002D0398 {ClMaroon},    {DevDryland}
      $00002BD5 {ClMaroon}, {UndDryland}
      ClGreen,  {Swamp}
      $00005500, {CLWhite,  }{CypressSwamp}  {Darker Green}
      ClLime,   {InlandFreshMarsh}
      $00A4FFA4,{TidalFreshMarsh} {Lighter Green}
      ClOlive,  {TransitionSaltmarsh} {Lighter Teal}
      $00A6A600,{Regularly flooded Saltmarsh }
      ClPurple, {Mangrove}
      $00B3FFFF,{EstuarineBeach}
      ClSilver, {TidalFlat}
      ClYellow, {OceanBeach}  {Lighter Yellow}
      $000059B3,{OceanFlat}   {Light Brown}
      $00FF80FF,{RockyIntertidal} {pink}
      $00FFA8A8,{InlandOpenWater} {light blue}
      ClBlue,   {RiverineTidalOpenWater}
      ClBlue,   {EstuarineWater}
      ClBlue,   {TidalCreek}
      ClNavy,   {OpenOcean}
      $004080FF,{Irregularly Flooded Brackish Marsh is Orange}
      $00006CD9,{Tall Spartina is dark orange, (Shows only when Requested)}
      $00004080,{Inland Shore is Brown}
      $00003E00,{Tidal Swamp is very dark green, almost black}
      ClWhite,   {ClWhite   {Blank}
      $0098BA8F, {Veg Tidal Flat is grey-green}
      $0075D6FF); {backshore is light orange, between dry land and beach}


  CLOpenWater = $00A74D00;
  CLLowTidal = $0000FFFF;
  CLSaltMarsh = $000000A8;
  CLTransitional = $0000FF55;
  CLFreshTidal = $00007326;
  ClFreshNonTidal = $00BEFFD2;
  ClUpland = $00B2B2B2;

{
  CLOpenWater = ClNavy;
  CLLowTidal = ClSilver;
  CLSaltMarsh = $00AA99;
  CLTransitional = $00DDFF;
  CLFreshTidal = $EEFF99;
  ClUpland = $00005400;
  ClFreshNonTidal = $0003D66C;   Colors pre 10-14-2011 }

  AlternativeGridColors: GridColorType
   = (ClUpland,{DevDryland}
      ClUpland,{UndDryland}
      ClFreshNonTidal,{Swamp}
      ClFreshNonTidal,{CypressSwamp}
      ClFreshNonTidal,{InlandFreshMarsh}
      CLFreshTidal,  {TidalFreshMarsh} {Lighter Green}
      CLTransitional,  {TransitionSaltmarsh} {Lighter Teal}
      CLSaltMarsh,  {Regularly flooded Saltmarsh }
      CLTransitional,  {Mangrove}
      CLLowTidal, {EstuarineBeach}
      CLLowTidal, {TidalFlat}
      CLLowTidal, {OceanBeach}  {Lighter Yellow}
      CLLowTidal, {OceanFlat}   {Light Brown}
      CLLowTidal, {RockyIntertidal} {pink}
      CLOpenWater,   {InlandOpenWater} {light blue}
      CLOpenWater,   {RiverineTidalOpenWater}
      CLOpenWater,   {EstuarineWater}
      CLOpenWater,   {TidalCreek}
      CLOpenWater,   {OpenOcean}
      CLTransitional,  {Irregularly Flooded Brackish Marsh is Orange}
      $0000000,{Tall Spartina is dark orange, (Shows only when Requested)}  // Irrelevant
      ClFreshNonTidal,{Inland Shore is Brown}
      CLFreshTidal,   {Tidal Swamp is very dark green, almost black}
      ClWhite,   {ClWhite   {Blank}
      CLLowTidal, {Veg Tidal Flat is grey-green}
      CLTransitional); {backshore is light orange, between dry land and beach}

  ElevationColors: ElevGridColorType
   = ($006644EE, {lowest elevation within tide range <MLLW}
      $006655EE,
      $006666EE,
      $006677EE,
      $006688EE,
      $006699EE {highest elevation around MHWS + 2}  );

Type
  IPCCScenarios  = (Scen_A1B,Scen_A1T,Scen_A1F1,Scen_A2,Scen_B1,Scen_B2);
  IPCCEstimates  = (Est_Min,Est_Mean,Est_Max);
  ProtScenario   = (NoProtect, ProtDeveloped, ProtAll);
  ESlammError    = Exception;

Const DiskUnit = 642190;

Const NUM_SAL_METRICS = 4;  {MLLW, MTL, MHW, 30D}

Type TSalStats = packed Record
   N: Array[1..NUM_SAL_METRICS, Category] of Integer;
   Min, Max, Sum, Sum_e2, Mean, Stdev, p05, p95: Array[1..NUM_SAL_METRICS,Category] of Double;
   Histogram : Array[1..NUM_SAL_METRICS,Category,0..(HistWidth-1)] of Integer;
 End;  

Const SliceIncrement = 0.1; {km}  {slice increment for estuary width accounting}

Type

  TDoubleArray = Array of Double;
  PDoubleArray = ^TDoubleArray;

  TSRecord = Record
               Year: Integer;
               Value: Double;
             End;

  PTSRArray = ^TSRArray;
  TSRArray  =  Array of TSRecord;

Const N_FTABLE = 60;
Type  TFTable  = Array[0..N_FTABLE-1] of Double; {Array of volume at depth relationships}

  TFWFlow     = Class
                   Name: String;
                   ExtentOnly   : Boolean;
                   UseTurbidities: Boolean;
                   SWPPT,FWPPT: Double;  // salinity of fresh water and salt water sources
                   Origin_km   : Double;  //origin of salt wedge, extent of freshwater flow in SLAMM river km

                   NTurbidities: Integer;
                   Turbidities: TSRArray;

                   NFlows: Integer;
                   MeanFlow: TSRArray;
                   ManningsN: Double;    // not currently used
                   FlWidth : Double;     // not currently used
                   InitSaltWedgeSlope: Double; {slope of the salt wedge, m/m}

                   TSElev  : Double; {Max. TS elev in meters above MTL}
                   {---------------- Calculations Below ---------------------}
                   RetTime: Array [1..NUM_SAL_METRICS] of Array of Single;  {Retention Time at MLLW, MTL, MHHW}
                   WaterZ: Array [1..NUM_SAL_METRICS] of Array of Single;   {Depth of Water at MLLW, MTL, MHHW}
                   XS_Salinity: Array [1..NUM_SAL_METRICS] of Array of Single;  {Weighted avg salinity of cross section (segment Rn) at MLLW, MTL, MHHW}

                   SubsiteArr: Array of TSubSite;
                   FTables: Array of TFTable;
                   D2Origin: Array of Single;
                   NumCells: Array of Integer;
                   Poly: TPolygon;
                   SW_Z_Mouth, SW_Z_Data: Double;

                   OriginArr, MouthArr: TPointArray;
                   NumSegments: Integer;

                   {CALCULATION VARS BELOW, No Loading or Saving Reqd.}
                   Plume, Vect: PDoubleArray;
                   EstuaryLength : Double;
                   RiverMouthIndex: Integer;
                   OriginLine, PlumeLine: Array of Tline;     // origin is perpendicular to line crossing start point, plume is pependicular to line crossing end point.
                   Midpts       : TPointArray;
                   MaxRn        : Integer;
                   TestMin, TestRange: Double; {range for F-Tables}

                   OceanSubsite : TSubsite;
                   RetentionInitialized: Boolean;

                   Constructor Create;
                   Function SaltWedgeSlopeByYear(Yr: Integer): Double;
                   Function FlowByYear(Yr: Integer): Double;
                   Function TurbidityByYear(Yr: Integer): Double;
                   Function ElevByVolume(VolWat: Double; RSeg: Integer): Double;
                   Procedure CellByRN(RSeg: Integer; Var OutR, OutC: Integer);
                   Function VolumeByElev(Elev: Double; RSeg: Integer): Double;
                   Function SaltHeight(TideHeight,SLR: Double; RSeg, Yr: Integer): Double;  {elevation of salt wedge relative to MTL}

                   Procedure Store(Var TS: TStream);
                   Constructor Load(ReadVersionNum : Double; TS: TStream);
                   Destructor Destroy;  override;
                 End; {TFWFlow}

  TFWFlows   = Array of TFWFlow;


CONST NUM_CAT_COMPRESS = 3;       {FIXME, Make Editable?}
TYPE
  PCompressedCell = ^CompressedCell;
  CompressedCell = packed RECORD
             Cats      : Array[1..NUM_CAT_COMPRESS] of Category;
             MinElevs  : Array[1..NUM_CAT_COMPRESS] of Single;
             Widths    : Array[1..NUM_CAT_COMPRESS] of single;
             TanSlope  : Single;
             ProtDikes : Boolean; {is protected by dikes?}
             ElevDikes : Boolean;  //consider optimizing memory by combining two booleans into a byte
             MaxFetch  : Single;
             Sal       : Array[1..NUM_SAL_METRICS] of Single;
             SalHeightMLLW : single; {meters}
             D2Channel : single;  {meters}
             Uplift    : single; {mm/year}
             DikeIndex : SmallInt;  {index to dike info record if estuarine diked}
             ErosionLoss, BTFErosionLoss: Single;
             MTLminusNAVD: single; {mtl minus navd88 for output purposes}
             D2MLLW, D2MHHW, ProbSAV : Single;
             ImpCoeff : Integer;
             SubsiteIndex: SmallInt;   // save subsite index to speed execution
//             MinFWFlowIndex: SmallInt; // save minimum FWFlow Index to speed execution
           End;

Const
  BLOCKSIZE = 10000;  {cells in one contiguous memory location}

Type
  DikeInfoRec = Packed Record
                  UpRow,UpCol: Integer;
                  Dist_Origin: Double;
                End;

  TDikeInfoArr = Array of DikeInfoRec;


  PMSArray2 = ^MSArray2;
  MSArray2 = Array[0..(BLOCKSIZE-1)] of CompressedCell;

  MSArray  = Array of PMSArray2;
{  MSFile  = FILE OF Cell_Rec; }


Const

  LabelProtect: Array[NoProtect..ProtAll] of String =
            ('No Protect','Protect Developed','Protect All');

  LabelIPCC : Array[Scen_A1B..Scen_B2] of String =
              ('Scenario A1B','Scenario A1T','Scenario A1F1','Scenario A2','Scenario B1','Scenario B2');

  LabelIPCCEst : Array[Est_Min..Est_Max] of String =
              ('Minimum','Mean','Maximum');

  LabelFixed: Array[1..3] of String {[31]} = ('1 meter','1.5 meter','2 meter');

Var  MinHist: Double = 0;  //salinity min
     MaxHist: Double = 35.0;  //salinity max


Type FractalRecord = Record
       NBox,
       MinFDRow,
       MinFDCol,
       MaxFDRow,
       MaxFDCol: Array[0..1] of Integer;
       FractalP,
       FractalD: Array[0..1] of Double;
       ShoreProtect: Double;
     End; {FractalRecord}

Procedure QuickSort(Var Arr: Array of Single; iLo, iHi: Integer);
Function CopyFile(Src,Dst: String): Boolean;

Var
   {Sensitivity Analysis Vars}
   SensAnalysis        : Boolean;
   UncAnalysis         : Boolean;      // Is there an uncertainty analysis running now?

   GlobalTS: TStream;

   TSText: Boolean;
   GlobalFile: TextFile;
   GlobalLN: Integer;

Procedure TSWrite(nm: String; i: Integer); overload;
Procedure TSWrite(nm: String; i: Double); overload;
Procedure TSWrite(nm: String; i: Boolean); overload;
Procedure TSWrite(nm: String; i: String); overload;

Procedure TSRead(nm: String; var i: Integer); overload;
Procedure TSRead(nm: String; var i: Double); overload;
Procedure TSRead(nm: String; var i: Boolean); overload;
Procedure TSRead(nm: String; var i: String; RVN: Double); overload;

Function LinearInterpolate(OldVal,NewVal,OldTime,NewTime,InterpTime: Double): Double;

Function IntInString(I: Integer; Str: String): Boolean;
Function AbbrString(InStr: String; DelimChar: Char):String;

Implementation

    Function AbbrString(InStr: String; DelimChar: Char):String;  // J.Guzman
    {Abridges a String to what appears before the DELIMCHAR }
    Var jj : integer;

    Begin
      jj := Pos(DelimChar,InStr);
      AbbrString := Copy(InStr,1,jj-1);
    End; {AbbrString}


   Function CheckName(Nm: String):String;
   Var NameStr, InStr: String;

   Begin
     Inc(GlobalLN);
     Readln(GlobalFile,InStr);
     NameStr := AbbrString(InStr,':');

     If NameStr <> Nm then Raise ESLAMMError.Create('Text Read Name Mismatch, Line '+IntToStr(GlobalLN)+
         ' Expecting Variable "'+Nm+'"; read variable "'+NameStr+'"');

     If Length(NameStr) >= Length(InStr)
       then Raise ESLAMMError.Create('No ":" or blank after ":", Reading Variable "'+Nm+'"  Line '+IntToStr(GlobalLN));

     Delete(InStr,1,Length(NameStr)+1);
     Result := InStr;
   End;

   Procedure TSWrite(nm: String;i: Integer); overload;
   Begin
     If TSText
       then Writeln(GlobalFile,nm,':',i)
       else GlobalTS.Write(i,Sizeof(i));

   End;

   Procedure TSWrite(nm: String;i: Double); overload;
   Begin
     If TSText
       then Writeln(GlobalFile,nm,':',i)
       else GlobalTS.Write(i,Sizeof(i));
   End;

   Procedure TSWrite(nm: String;i: Boolean); overload;
   Begin
     If TSText
       then Writeln(GlobalFile,nm,':',i)
       else GlobalTS.Write(i,Sizeof(i));
   End;

   Procedure TSWrite(nm: String;i: String); overload;
   Var SS: Integer;
   Begin
     If TSText
       then Writeln(GlobalFile,nm,':',i)
       else
         Begin
           SS := Length(i);
           GlobalTS.Write(SS,Sizeof(Integer));
           If SS>0 then GlobalTS.Write(Pointer(i)^,SS*Sizeof(Char));
         End;
   End;

   Procedure TSRead(nm: String;Var i: Integer); overload;
   Begin
     If TSText
       then Try
              i := StrToInt(CheckName(nm));
            Except
              Raise ESLAMMError.Create('Integer Conversion Error, Line '+IntToStr(GlobalLN));
            End
       else GlobalTS.Read(i,Sizeof(i));
   End;

   Procedure TSRead(nm: String;Var i: Double); overload;
   Begin
     If TSText
       then Try
              i := StrToFloat(CheckName(nm));
            Except
              Raise ESLAMMError.Create('Integer Conversion Error, Line '+IntToStr(GlobalLN));
            End
       else GlobalTS.Read(i,Sizeof(i));
   End;

   Procedure TSRead(nm: String;Var i: Boolean); overload;
   Begin
     If TSText
       then Try
              i := Pos('TRUE',Uppercase(CheckName(nm)))>0;
            Except
              Raise ESLAMMError.Create('Integer Conversion Error, Line '+IntToStr(GlobalLN));
            End
       else GlobalTS.Read(i,Sizeof(i));
   End;

   Procedure TSRead(nm: String;Var i: String; RVN: Double); overload;
   Var SS: Integer;  {String Size}
       Temp: AnsiString;
   Begin
     If TSText
       then i := CheckName(nm)
       else Begin
              GlobalTS.Read(SS,Sizeof(Integer));
              If SS > 9999 then Raise ESLAMMError.Create('ReadStr Error, Input File Corrupted');
              Setlength(i,SS);
              if RVN < 6.495
                then
                  Begin
                   SetLength(Temp, SS);              // <<-- Use temporary AnsiString
                   GlobalTS.Read(Pointer(Temp)^, SS * SizeOf(AnsiChar));  // <<-- Specify buffer size in bytes
                   i := string(Temp);                // <<-- Widen string to Unicode
                  End
                else
                  If SS>0 then GlobalTS.Read(Pointer(i)^,SS*SizeOf(Char));
            End;
   End;

{--------------------------------------------------------------------}
  Function LinearInterpolate(OldVal,NewVal,OldTime,NewTime,InterpTime: Double): Double;
  {Interpolates to InterpTime between two points, OldPoint and NewPoint}
  Begin
    If (InterpTime>NewTime) or (InterpTime<OldTime)                      {Test Input}
       then Raise ESLAMMERROR.Create('Interpolation Timestamp Error');

    LinearInterpolate := OldVal + ((NewVal-OldVal)/(NewTime-OldTime)) * (InterpTime-OldTime);
                         { y1 }   {             Slope  (dy/dx)      }   {    Delta X       }
  End;
{--------------------------------------------------------------------}
  Procedure QuickSort(Var Arr: Array of Single; iLo, iHi: Integer);
  Var
    Lo, Hi : Integer;
    T, Mid : Double;
  Begin
    Lo := iLo;
    Hi := iHi;
    Mid := Arr[(Lo + Hi) div 2];
    repeat
      while Arr[Lo] < Mid do Inc(Lo);
      while Arr[Hi] > Mid do Dec(Hi);
      if Lo <= Hi then
      begin
        T := Arr[Lo];
        Arr[Lo] := Arr[Hi];
        Arr[Hi] := T;
        Inc(Lo);
        Dec(Hi);
      end;
    until Lo > Hi;
    if Hi > iLo then QuickSort(Arr, iLo, Hi);
    if Lo < iHi then QuickSort(Arr, Lo, iHi);
  End;
{--------------------------------------------------------------------}

Constructor TFWFlow.Create;
Var i: Integer;
Begin
  Name := 'New Flow';
  ExtentOnly := False;
  UseTurbidities := False;
  FWPPT := 0;
  SWPPT := 30;
  Origin_km := 0;
  NTurbidities := 1;
  SetLength(Turbidities,10);
  Turbidities[0].Year := 2000;
  Turbidities[0].Value := 0;

  NFlows := 0;
  MeanFLow := Nil;
  ManningsN := 0;
  FLWidth := 0;
  InitSaltWedgeSlope:= 0;

  Poly := tpolygon.create;
  OriginArr := nil;
  MouthArr := nil;
  Plume := nil;
  Vect := nil;

  NumSegments := 0;

  MidPts := nil;

  D2Origin := nil;

  For i := 1 to 3 do
   Begin
    RetTime[i] := nil;
    WaterZ[i] := nil;
    XS_Salinity[i] := nil;
   End;

  OriginLine := nil;
  PlumeLine := nil;
  MaxRn := 0;
  TestMin := 0; TestRange := 0;

  OceanSubsite := nil;
  SubsiteArr := nil;
  RetentionInitialized := False;

  inherited;
End;

{*************************************************************************
 * FUNCTION:   CCW (CounterClockWise)
 *
 * PURPOSE
 * Determines, given three points, if when travelling from the first to
 * the second to the third, we travel in a counterclockwise direction.
 *
 * RETURN VALUE
 * (int) 1 if the movement is in a counterclockwise direction, -1 if not.
 *************************************************************************}


function CCW(Var p0,p1,p2:TPoint): Integer;
var
  dx1, dx2: LongInt;
  dy1, dy2: LongInt;
begin
  dx1 := p1.x - p0.x ;
  dx2 := p2.x - p0.x ;
  dy1 := p1.y - p0.y ;
  dy2 := p2.y - p0.y ;

  {* This is basically a slope comparison: we don't do divisions because
   * of divide by zero possibilities with pure horizontal and pure
   * vertical lines.  *}

  if((dx1 * dy2) > (dy1 * dx2))then
    Result := 1
  else
    Result := -1;
end;


{*************************************************************************
 * FUNCTION:   Intersect
 *
 * PURPOSE
 * Given two line segments, determine if they intersect.
 *
 * RETURN VALUE
 * TRUE if they intersect, FALSE if not.
 *************************************************************************}


function Intersect(Var p1,p2,p3,p4: TPoint): Boolean;
begin
  Result := (((CCW(p1, p2, p3) * CCW(p1, p2, p4)) <= 0)and
            (( CCW(p3, p4, p1) * CCW(p3, p4, p2)  <= 0))) ;
end;



Constructor TPolygon.Create;
Begin
  NumPts := 0;
  svMinRow := -9999;
  svMaxRow := -9999;
  svMinCol := -9999;
  svMaxCol := -9999;
  TPoints := nil;
End;

Constructor TPolygon.Load(ReadVersionNum: Double; TS: TStream; ReadHeader: Boolean);
Var i: Integer;
Begin
  If ReadHeader then
    Begin
      TSRead('Poly.NumPts',NumPts);
      SetLength(TPoints,NumPts);
    End;

  For i := 1 to NumPts do
    Begin
      TSRead('TPoints[i-1].x',TPoints[i-1].x);
      TSRead('TPoints[i-1].y',TPoints[i-1].y);
    End;

  svMinRow := -9999;
  svMaxRow := -9999;
  svMinCol := -9999;
  svMaxCol := -9999;
End;

Procedure TPolygon.Store;
Var i: Integer;
Begin
  TSWrite('Poly.NumPts',NumPts);
  For i := 1 to NumPts do
    Begin
      TSWrite('TPoints[i-1].x',TPoints[i-1].x);
      TSWrite('TPoints[i-1].y',TPoints[i-1].y);
    End;
End;



Destructor TPolygon.Destroy;
Begin
  TPoints := nil;
  inherited;
End;

Function TPolygon.MinRow: Integer;
Var i,MR: Integer;
Begin
  If svMinRow <> -9999
    then MinRow := svMinRow
    else
      Begin
        MR := High(Integer);
        For i := 0 to NumPts-1 do
          If TPoints[i].y < MR then MR := TPoints[i].y;
        MinRow := MR;
        svMinRow := MR;
      End;
End;

Function TPolygon.MaxRow: Integer;
Var i,MR: Integer;
Begin
  If svMaxRow <> -9999
    then MaxRow := svMaxRow
    else
      Begin
        MR := Low(Integer);
        For i := 0 to NumPts-1 do
          If TPoints[i].y > MR then MR := TPoints[i].y;
        MaxRow := MR;
        svMaxRow := MR;
      End;
End;

Function TPolygon.MinCol: Integer;
Var i,MC: Integer;
Begin
  If svMinCol <> -9999
    then MinCol := svMinCol
    else
      Begin
        MC := High(Integer);
        For i := 0 to NumPts-1 do
          If TPoints[i].x < MC then MC := TPoints[i].x;
        MinCol := MC;
        svMinCol := MC;
      End;
End;

Function TPolygon.MaxCol: Integer;
Var i,MC: Integer;
Begin
  If svMaxCol <> -9999
    then MaxCol := svMaxCol
    else
      Begin
        MC := Low(Integer);
        For i := 0 to NumPts-1 do
          If TPoints[i].x > MC then MC := TPoints[i].x;
        MaxCol := MC;
        svMaxCol := MC;
      End;
End;


Function TPolygon.InPoly(Row,Col: Integer):Boolean;
Var ThisP: TPoint;

{*************************************************************************
 * FUNCTION:   G_PtInPolygon
 *
 * PURPOSE
 * This routine determines if the point passed is in the polygon. It uses
 * the classical polygon hit-testing algorithm: a horizontal ray starting
 * at the point is extended infinitely rightwards and the number of
 * polygon edges that intersect the ray are counted. If the number is odd,
 * the point is inside the polygon.
 *
 *************************************************************************}

    {*************************************************************************
     * FUNCTION:   PtInPolyRect
     *
     * PURPOSE
     * This routine determines if a point is within the smallest rectangle
     * that encloses a polygon.  Optimized 9/25/2012 JSC
     *
     *************************************************************************}

{    function PtInPolyRect(ptTest: TPoint; var prbound: TRect): Boolean;
    var
      pt: TPoint;
      i: Word;
    begin
       prbound := Rect(MinCol, MinRow, MaxCol, MaxRow);
       Result := PtInRect(prbound,ptTest);
    end; }

    function PtInPolyRect: Boolean;
    begin
       Result := (Row>=MinRow) and (Row<=MaxRow) and
                 (Col>=MinCol) and (Col<=MaxCol);
    end;

    Function G_PtInPolygon:Boolean;
    var
      i: Integer;
      pt2: TPoint;
      wnumintsct: Word;
      prbound: TRect;
    begin
      Result := False;
      if not PtInPolyRect then Exit;

      wnumintsct := 0;
      pt2 := ThisP;
      pt2.x := MaxCol+50;

      // Now go through each of the lines in the polygon and see if it intersects
      for i := 0 to High(TPoints)-1 do
        if(Intersect(ThisP,pt2,TPoints[i],TPoints[i+1]))then
          Inc(wnumintsct);
      // And the last line
      if(Intersect(ThisP, pt2, TPoints[High(TPoints)],TPoints[0]))then
         Inc(wnumintsct);
      // If wnumintsct is odd then the point is inside the polygon
      Result := Odd(wnumintsct);
    end;

Begin
  InPoly := False;
  If TPoints = nil then Exit;
  If Length(TPoints) > NumPts then SetLength(TPoints,NumPts);
  ThisP.X := Col; ThisP.Y := Row;
  InPoly := G_PtInPolygon;
End;

Constructor TFWFlow.Load(ReadVersionNum : Double; TS: TStream);
var i: integer;
    Junk: Double;
Begin
  TSRead('Name',Name,ReadVersionNum);
  If ReadVersionNum < 6.025 then TSRead('Junk',Junk);
  Poly := TPolygon.Create;

  If ReadVersionNum < 6.375 then
    Begin
      TSRead('Poly.NumPts',Poly.NumPts);
      SetLength(Poly.TPoints,Poly.NumPts);
    End;

  If ReadVersionNum > 6.025 then
    Begin
      TSRead('UseTurbidities',UseTurbidities);

      if ReadVersionNum > 6.265 then
        Begin
          TSRead('SWPPT',SWPPT);
          TSRead('FWPPT',FWPPT);
          If ReadVersionNum > 6.305
            then TSRead('Origin_KM',Origin_KM)
            else Begin
                   TSRead('SWedge_Z',Origin_KM);
                   Origin_KM := 0;
                 End;
        End
      else
        Begin
          TSRead('SalPPT',FWPPT);   //SalPPT was the old variable name for fresh water ppt
          SWPPT := 30;
          Origin_KM := 0;
        End;

      TSRead('NTurbidities',NTurbidities);
      SetLength(Turbidities,NTurbidities);
      For i := 0 to NTurbidities-1 do
        Begin
          TSRead('Turbidities[i].Year',Turbidities[i].Year);
          TSRead('Turbidities[i].Value',Turbidities[i].Value);
        End;
      TSRead('NFlows',NFlows);
      SetLength(MeanFlow,NFlows);
      For i := 0 to NFlows-1 do
        Begin
          TSRead('MeanFlow[i].Year',MeanFlow[i].Year);
          TSRead('MeanFlow[i].Value',MeanFlow[i].Value);
        End;
      TSRead('ManningsN',ManningsN);
      TSRead('FlWidth',FlWidth);
      TSRead('SaltWedgeSlope',InitSaltWedgeSlope);
    End
  else  {readversionNum <= 6.025}
      Begin
        UseTurbidities := True;
        FWPPT := 0;
        SWPPT := 30;
        Origin_KM := 0;
        NTurbidities := 1;
        SetLength(Turbidities,NTurbidities);
        TSRead('Turbidities[0].Value',Turbidities[0].Value);
        Turbidities[0].Year := 2000;
        NFlows := 0;
        MeanFlow := Nil;
        ManningsN := 0;
        FlWidth := 0;
        InitSaltWedgeSlope := 1e-5;
      End;

  If ReadVersionNum > 6.205
     then TSRead('ExtentOnly',ExtentOnly)
     else ExtentOnly := False;

  TSRead('TSElev',TSElev);

  Poly.Load(ReadVersionNum, TS, ReadVersionNum>6.375);

  TSRead('NumSegments',NumSegments);
  SetLength(OriginArr,NumSegments);
  SetLength(MouthArr,NumSegments);

  For i := 1 to NumSegments do
    Begin
      TSRead('OriginArr[i-1].x',OriginArr[i-1].x);
      TSRead('OriginArr[i-1].y',OriginArr[i-1].y);
      TSRead('MouthArr[i-1].x',MouthArr[i-1].x);
      TSRead('MouthArr[i-1].y',MouthArr[i-1].y);
    End;

  TSRead('SW_Z_Data',SW_Z_Data);
  TSRead('SW_Z_Mouth',SW_Z_Mouth);

  If ReadVersionNum < 6.025 then
    Begin
      TSRead('Junk',Junk {Salt_Brack});     {Frac Fresh Bounds}
      TSRead('Junk',Junk {Brack_TFresh});
      TSRead('Junk',Junk {TFresh_TSwamp});
    End;

  Plume := nil;
  Vect := nil;
  MidPts := nil;

  D2Origin := nil;
  For i := 1 to 3 do
   Begin
    RetTime[i] := nil;
    WaterZ[i] := nil;
    XS_Salinity[i] := nil;
   End;

  OriginLine := nil;
  PlumeLine := nil;
  MaxRn := 0;
  TestMin := 0; TestRange := 0;

  OceanSubsite := nil;
  RetentionInitialized := False;
End;


function TFWFlow.SaltHeight(TideHeight, SLR: Double; RSeg, Yr: Integer): Double;
Var MRn: Double;
    OR2: Double;
begin
  OR2 := ORIGIN_KM;
  If OR2 < 0.05 then OR2 := MaxRn * SliceIncrement;

  OR2 := OR2 - ((TideHeight+SLR)* 4.82);  // JSC 7/27/2012  Relationship of 4.82 km on salt wedge origin per meter of tide height derived from GA data.
                    {m}     {m}
  Mrn := (OR2/SliceIncrement);
          {km}    {km/n}

  SaltHeight := TideHeight -(Mrn - RSeg) * SliceIncrement  * 1000 * SaltWedgeSlopeByYear(Yr) ;
  {m to mtl}      {m}        {n}   {n}        {km}          {m/km}      {m/m}
end;

Procedure TFWFlow.Store;
var i: integer;
Begin
  TSWrite('Name',Name);

  TSWrite('UseTurbidities',UseTurbidities);
  TSWrite('SWPPT',SWPPT);
  TSWrite('FWPPT',FWPPT);
  TSWrite('Origin_KM',Origin_KM);

  TSWrite('NTurbidities',NTurbidities);
  For i := 0 to NTurbidities-1 do
    Begin
      TSWrite('Turbidities[i].Year',Turbidities[i].Year);
      TSWrite('Turbidities[i].Value',Turbidities[i].Value);
    End;
  TSWrite('NFlows',NFlows);
  For i := 0 to NFlows-1 do
    Begin
      TSWrite('MeanFlow[i].Year',MeanFlow[i].Year);
      TSWrite('MeanFlow[i].Value',MeanFlow[i].Value);
    End;
  TSWrite('ManningsN',ManningsN);
  TSWrite('FlWidth',FlWidth);

  TSWrite('SaltWedgeSlope',InitSaltWedgeSlope);

  TSWrite('ExtentOnly',ExtentOnly);
  TSWrite('TSElev',TSElev);

  Poly.Store;

  TSWrite('NumSegments',NumSegments);
  For i := 1 to NumSegments do
    Begin
      TSWrite('OriginArr[i-1].x',OriginArr[i-1].x);
      TSWrite('OriginArr[i-1].y',OriginArr[i-1].y);
      TSWrite('MouthArr[i-1].x',MouthArr[i-1].x);
      TSWrite('MouthArr[i-1].y',MouthArr[i-1].y);
    End;

  TSWrite('SW_Z_Data',SW_Z_Data);
  TSWrite('SW_Z_Mouth',SW_Z_Mouth);

End;


function TFWFlow.SaltWedgeSlopeByYear(Yr: Integer): Double;
Var DeltaFlow: Double;
Begin
  DeltaFlow := FlowByYear(Yr) - FlowByYear(0);        {cfs}
  SaltWedgeSlopeByYear := InitSaltWedgeSlope + DeltaFlow * 2.8E-7;
               {relationship of 2.8E-7 in slope per CFS derived from four GA rivers}

End;


function TFWFlow.FlowByYear(Yr: Integer): Double;

Var PReadArr : PTSRArray;
    PCount : PInteger;
    i: Integer;
    Found: Boolean;
begin
  Result := 0; {not used}

  PCount := @NFlows;
  PReadArr := @MeanFlow;

  If PCount^ = 0 then Raise ESLAMMError.Create('Flow Time-Series Data Missing.');

  Found := False;
  If (PCount^=1) or (Yr < PReadArr^[0].Year)
    then Begin
           Result := PReadArr^[0].Value;
           Found := True;
         End
    else For i := 0 to PCount^-1 do
      If (Yr=PReadArr^[i].Year)
        then Begin
               Result := PReadArr^[i].Value;
               Found := True;
               Break;
             End
        else if (Yr < PReadArr^[i].Year)
           then Begin
                  Result := LinearInterpolate(PReadArr^[i-1].Value,PReadArr^[i].Value,
                                              PReadArr^[i-1].Year,PReadArr^[i].Year,Yr);
                  Found := True;
                  Break;
                End;

  If Not Found then Result := PReadArr^[PCount^-1].Value;  {year greater than all data, choose last point}
end;



function TFWFlow.TurbidityByYear(Yr: Integer): Double;

Var PReadArr : PTSRArray;
    PCount : PInteger;
    i: Integer;
    Found: Boolean;
begin
  Result := 1.0;
  If not UseTurbidities then  Exit;

  PReadArr := @Turbidities;
  PCount := @NTurbidities;

  If PCount^ = 0 then Raise ESLAMMError.Create('Turbidity Time-Series Data Missing.');

  Found := False;
  If (PCount^=1) or (Yr < PReadArr^[0].Year)
    then Begin
           Result := PReadArr^[0].Value;
           Found := True;
         End
    else For i := 0 to PCount^-1 do
      If (Yr=PReadArr^[i].Year)
        then Begin
               Result := PReadArr^[i].Value;
               Found := True;
               Break;
             End
        else if (Yr < PReadArr^[i].Year)
           then Begin
                  Result := LinearInterpolate(PReadArr^[i-1].Value,PReadArr^[i].Value,
                                              PReadArr^[i-1].Year,PReadArr^[i].Year,Yr);
                  Found := True;
                  Break;
                End;

  If Not Found then Result := PReadArr^[PCount^-1].Value;  {year greater than all data, choose last point}
end;


Destructor TFWFlow.Destroy;
Var i: Integer;
Begin
  FTables := nil;
  SubsiteArr := nil;
  For i := 1 to 3 do
   Begin
    RetTime[i] := nil;
    WaterZ[i] := nil;
    XS_Salinity[i] := nil;
   End;

  NumCells := nil;
  D2Origin := nil;

  Plume := nil;
  Vect := nil;
  MidPts := nil;
  OriginLine := nil;
  PlumeLine := nil;

  inherited;
End;

Procedure TFWFlow.CellByRN(RSeg: Integer; Var OutR, OutC: Integer);
Var i: Integer;
    Distance: Double;
    Found:Boolean;
    FracLine: Double;
Begin
  Distance := RSeg*SliceIncrement;
  Found := False;
  for i := 1 to numsegments -1 do
   If Distance <= D2Origin[i] then
     Begin
       FracLine := (Distance - D2Origin[i-1])/(D2Origin[i]-D2Origin[i-1]);  // fraction along line of RSeg
       OutC := OriginArr[i-1].X + Round((MouthArr[i-1].X-OriginArr[i-1].X)*FracLine);
       OutR := OriginArr[i-1].Y + Round((MouthArr[i-1].Y-OriginArr[i-1].Y)*FracLine);
       Found := True;
       Break;
     End;

     If Not Found then   //extrapolate into plume
       Begin
         FracLine := (Distance - D2Origin[numsegments-1])/(EstuaryLength - D2Origin[numsegments-1]);  // fraction along line of RSeg
         OutC := OriginArr[numsegments-1].X + Round((MouthArr[numsegments-1].X-OriginArr[numsegments-1].X)*FracLine);
         OutR := OriginArr[numsegments-1].Y + Round((MouthArr[numsegments-1].Y-OriginArr[numsegments-1].Y)*FracLine);
       End;

End;


function TFWFlow.ElevByVolume(VolWat: Double; RSeg: Integer): Double;
Var MinElev,MaxElev, MinVol, MaxVol: Double;
    i: Integer;
begin
  Result := 0; {not used}
  If FTables[RSeg][0] > VolWat  {water is all below test interval, assign lowest elevation tested}
       then Result := TestMin
       else If FTables[RSeg][N_FTable-1] < VolWat
         then Result := TestMin + TestRange {water does not fit in all tested regions, assign highest elev tested}
         else For i := 1 to N_FTable-1 do
           if FTables[RSeg][i] > VolWat then
             Begin
               MinElev :=  TestMin + (TestRange / (N_FTable-1))*(i-1);
               MaxElev :=  TestMin + (TestRange / (N_FTable-1))*i;
               MinVol := FTables[RSeg][i-1];
               MaxVol := FTables[RSeg][i];
               Result := LinearInterpolate(MinElev,MaxElev,MinVol,MaxVol,VolWat);
               Break;
             End; {if}
end;

function TFWFlow.VolumeByElev(Elev: Double; RSeg: Integer): Double;
Var FTHeight: Double;
    MinElev,MaxElev, MinVol, MaxVol: Double;
    i: Integer;
begin
  Result := 0; {not used}
  If (Elev<TestMin)
    then Result := FTables[RSeg][0]  {elev below minimum in table, assign lowest volume}
    else if (Elev>TestMin+TestRange)
      then Result := FTables[RSeg][N_FTable-1] {elev above maximum in table, assign highest volume}
      else For i := 1 to N_FTable-1 do
        Begin
          FTHeight := TestMin + (TestRange / (N_FTable-1))*i;
          If Elev<FTHeight then
            Begin
               MinElev :=  TestMin + (TestRange / (N_FTable-1))*(i-1);
               MaxElev :=  FTHeight;
               MinVol := FTables[RSeg][i-1];
               MaxVol := FTables[RSeg][i];
               Result := LinearInterpolate(MinVol,MaxVol,MinElev,MaxElev,Elev);
               Break;
            End;
        End;
end;

{-----------------------------------------------------------------}

Constructor TSite.Create;
Begin
  SubSites := nil;
  OutputSites := Nil;
  ROSBounds := Nil;
  GlobalSite := TSubSite.Create;
End;

Destructor TSite.Destroy;
Var i: Integer;
Begin
  For i := 0 to NSubSites-1 do
    SubSites[i].Destroy;

  GlobalSite.Destroy;

  SubSites := nil;
  For i := 0 to NOutputSites-1 do
    OutputSites[i].Poly.Free; 

  OutputSites := Nil;
  ROSBounds := Nil;

  inherited;
End;


Procedure TSite.LoadSubSite(ReadVersionNum: Double; TS: TStream; Var SS: TSubSite);

Var i, Freq_BigStormInt: Integer;
    OffshoreText: String;
    OffshoreChar: Char;

Begin
 SS := TSubSite.Create;

 With SS do
  Begin
    Poly.Load(ReadVersionNum, TS,True);

     TSRead('Description',Description,ReadVersionNum);
//      TSRead('Depth',Depth);
      TSRead('NWI_Photo_Date',NWI_Photo_Date);
      TSRead('DEM_date',DEM_date);

      If TSText then
          Begin
            TSRead('Direction_Offshore',OffShoreText,ReadVersionNum);
            OffshoreText := Uppercase(OffShoreText);
            OffShoreChar := OffShoreText[1];
            Case OffShoreChar of
              'N' : Direction_OffShore := Northerly;
              'S' : Direction_OffShore := Southerly;
              'E' : Direction_OffShore := Easterly;
              else  Direction_OffShore := Westerly;
            End; {Case}
          End
        else TS.Read(Direction_OffShore, Sizeof(Direction_Offshore));

      TSRead('Historic_trend',Historic_trend);
      TSRead('NAVD88MTL_correction',NAVD88MTL_correction);
      TSRead('GTideRange',GTideRange);
      TSRead('SaltElev',SaltElev);
      TSRead('MarshErosion',MarshErosion);
      TSRead('SwampErosion',SwampErosion);
      TSRead('TFlatErosion',TFlatErosion);
      TSRead('FixedRegFloodAccr',FixedRegFloodAccr);
      TSRead('FixedIrregFloodAccr',FixedIrregFloodAccr);
      TSRead('FixedTideFreshAccr',FixedTideFreshAccr);

      IF ReadVersionNum > 6.245 then
        Begin
          TSRead('InlandFreshAccr',InlandFreshAccr);
          TSRead('MangroveAccr',MangroveAccr);
          TSRead('TSwampAccr', TSwampAccr);
          TSRead('SwampAccr',SwampAccr);
        End
      Else
        Begin
          InlandFreshAccr := FixedTideFreshAccr;
          MangroveAccr := 7.0; {default}
          TSwampAccr := 1.1 ;  {default}
          SwampAccr := 0.3 ;   {default}
        End;

      TSRead('Fixed_TF_Beach_Sed',Fixed_TF_Beach_Sed);
      If ReadVersionNum < 6.255
        then Begin
               TSRead('Freq_BigStorm',Freq_BigStormInt);
               Freq_BigStorm := Freq_BigStormInt;
             End
        else TSRead('Freq_BigStorm',Freq_BigStorm);

      TSRead('Use_Preprocessor',Use_Preprocessor);

      TSRead('OW_Max_Width_Overwash',OW_Max_Width_Overwash);
      TSRead('OW_Beach_to_Ocean',OW_Beach_to_Ocean);
      TSRead('OW_Dryland_to_Beach',OW_Dryland_to_Beach);
      TSRead('OW_Est_to_Beach',OW_Est_to_Beach);
      TSRead('OW_Marsh_Pct_Loss',OW_Marsh_Pct_Loss);
      TSRead('OW_Mang_Pct_Loss',OW_Mang_Pct_Loss);

      For i := 0 to N_ACCR_MODELS-1 do
       If ReadVersionNum < 6.025
        then
          Begin
            UseAccrModel[i] := False;  {boolean}
            MaxAccr[i] := 0;  {mm/year}
            MinAccr[i] := 0; {mm/year}
            AccrA[i] := 0; {unitless}
            AccrB[i] := 0; {unitless}
            AccrC[i] := 0; {unitless}
            DistEffectMax[i] := 0; {meters}
            Dmin[i] := 1; {unitless scaling factor, set to 1.0 to ignore distance factor}
            SalinityTMax[i] := 0; {ppt}
            STMaxZone[i] := 0; {ppt}
            SNonTMax[i] := 1; {unitless multiplier, set to one to ignore salinity}
            AccrNotes[i] := '';
          End
        else
          Begin
            TSRead('UseAccrModel[i]',UseAccrModel[i]);
            TSRead('MaxAccr[i]',MaxAccr[i]);
            TSRead('MinAccr[i]',MinAccr[i]);
            TSRead('AccrA[i]',AccrA[i]);
            TSRead('AccrB[i]',AccrB[i]);
            If ReadVersionNum > 6.035
              then TSRead('AccrC[i]',AccrC[i])
              else AccrC[i] := 0;
            TSRead('DistEffectMax[i]',DistEffectMax[i]);
            TSRead('Dmin[i]',Dmin[i]);
            TSRead('SalinityTMax[i]',SalinityTMax[i]);
            TSRead('STMaxZone[i]',STMaxZone[i]);
            TSRead('SNonTMax[i]',SNonTMax[i]);
            TSRead('AccrNotes[i]',AccrNotes[i],ReadVersionNum);
          End;
  End;
End;

Procedure TSite.StoreSubSite(Var TS: TStream; Var SS: TSubSite);
Var i: Integer;
Begin
 With SS do
  Begin
    Poly.Store;

      TSWrite('Description',Description);
//    TSWrite('Depth',Depth);
      TSWrite('NWI_Photo_Date',NWI_Photo_Date);
      TSWrite('DEM_date',DEM_date);

      If TSText then
          Begin
            Case Direction_OffShore of
              Northerly : TSWrite('Direction_Offshore','Northerly');
              Southerly : TSWrite('Direction_Offshore','Southerly');
              Easterly : TSWrite('Direction_Offshore','Easterly');
              else  TSWrite('Direction_Offshore','Westerly');
            End; {Case}
          End
        else TS.Write(Direction_OffShore, Sizeof(Direction_Offshore));

      TSWrite('Historic_trend',Historic_trend);
      TSWrite('NAVD88MTL_correction',NAVD88MTL_correction);
      TSWrite('GTideRange',GTideRange);
      TSWrite('SaltElev',SaltElev);
      TSWrite('MarshErosion',MarshErosion);
      TSWrite('SwampErosion',SwampErosion);
      TSWrite('TFlatErosion',TFlatErosion);
      TSWrite('FixedRegFloodAccr',FixedRegFloodAccr);
      TSWrite('FixedIrregFloodAccr',FixedIrregFloodAccr);
      TSWrite('FixedTideFreshAccr',FixedTideFreshAccr);

      TSWrite('InlandFreshAccr',InlandFreshAccr);  // New 5/20/2011
      TSWrite('MangroveAccr',MangroveAccr);
      TSWrite('TSwampAccr', TSwampAccr);
      TSWrite('SwampAccr',SwampAccr);

      TSWrite('Fixed_TF_Beach_Sed',Fixed_TF_Beach_Sed);
      TSWrite('Freq_BigStorm',Freq_BigStorm);
      TSWrite('Use_Preprocessor',Use_Preprocessor);

      TSWrite('OW_Max_Width_Overwash',OW_Max_Width_Overwash);
      TSWrite('OW_Beach_to_Ocean',OW_Beach_to_Ocean);
      TSWrite('OW_Dryland_to_Beach',OW_Dryland_to_Beach);
      TSWrite('OW_Est_to_Beach',OW_Est_to_Beach);
      TSWrite('OW_Marsh_Pct_Loss',OW_Marsh_Pct_Loss);
      TSWrite('OW_Mang_Pct_Loss',OW_Mang_Pct_Loss);

      For i := 0 to N_ACCR_MODELS-1 do
        Begin
          TSWrite('UseAccrModel[i]',UseAccrModel[i]);
          TSWrite('MaxAccr[i]',MaxAccr[i]);
          TSWrite('MinAccr[i]',MinAccr[i]);
          TSWrite('AccrA[i]',AccrA[i]);
          TSWrite('AccrB[i]',AccrB[i]);
          TSWrite('AccrC[i]',AccrC[i]);
          TSWrite('DistEffectMax[i]',DistEffectMax[i]);
          TSWrite('Dmin[i]',Dmin[i]);
          TSWrite('SalinityTMax[i]',SalinityTMax[i]);
          TSWrite('STMaxZone[i]',STMaxZone[i]);
          TSWrite('SNonTMax[i]',SNonTMax[i]);
          TSWrite('AccrNotes[i]',AccrNotes[i]);
        End;

  End;
End;

Constructor TSite.Load(ReadVersionNum: Double; TS: TStream);
Var  i: Integer;
Begin
  TSRead('Rows',Rows);
  TSRead('Cols',Cols);
  TSRead('LLXCorner',LLXCorner);
  TSRead('LLYCorner',LLYCorner);
  TSRead('Scale',Scale);

  TSRead('NSubSites',NSubSites);
  SetLength(SubSites,NSubSites);
  For i := 0 to NSubSites-1 do
      LoadSubSite(ReadVersionNum,TS,Subsites[i]);

  TSRead('NOutputSites',NOutputSites);

  SetLength(OutputSites,NOutputSites);
  For i := 0 to NOutputSites-1 do
     With OutputSites[i] do
      Begin
        UsePolygon := False;
        If ReadVersionNum > 6.355 then TSRead('UsePolygon',UsePolygon);
        If UsePolygon then Poly := TPolygon.Load(ReadVersionNum, TS,True);
        TSRead('X1',Rec.X1);
        TSRead('Y1',Rec.Y1);
        TSRead('X2',Rec.X2);
        TSRead('Y2',Rec.Y2);
        TSRead('Description',Description,ReadVersionNum);
      End;

  LoadSubSite(ReadVersionNum,TS,GlobalSite);
End;


Procedure TSite.Store(Var TS: TStream);
Var  i: Integer;
Begin
  TSWrite('Rows',Rows);
  TSWrite('Cols',Cols);
  TSWrite('LLXCorner',LLXCorner);
  TSWrite('LLYCorner',LLYCorner);
  TSWrite('Scale',Scale);

  TSWrite('NSubSites',NSubSites);
{  SetLength(SubSites,NSubSites); }
  For i := 0 to NSubSites-1 do
      StoreSubSite(TS,Subsites[i]);

  TSWrite('NOutputSites',NOutputSites);
  For i := 0 to NOutputSites-1 do
     With OutputSites[i] do
      Begin
        TSWrite('UsePolygon',UsePolygon);
        If UsePolygon then Poly.Store;
        TSWrite('X1',Rec.X1);
        TSWrite('Y1',Rec.Y1);
        TSWrite('X2',Rec.X2);
        TSWrite('Y2',Rec.Y2);
        TSWrite('Description',Description);
      End;

  StoreSubSite(TS,GlobalSite);
End;

Function TSite.InOutSite(X {Col},Y {Row}:Integer; SiteNum: Integer): Boolean;  {Is this x,y in the output site?}
Begin
   With OutputSites[SiteNum-1] do
     If UsePolygon
       then Result := Poly.InPoly(Y,X)
       else With Rec do Result := (X>= Min(X1,X2)) and (X<=Max(X1,X2)) and (Y>=Min(Y1,Y2)) and (Y<=Max(Y1,Y2));
End;


Function TSite.GetSubSiteNum(X,Y:Integer; PC: pointer): Integer;

Begin
  if PCompressedCell(PC).SubsiteIndex <> -9999
    then Result := PCompressedCell(PC).SubsiteIndex
    else Begin
           Result := GetSubSiteNum(X,Y);
           PCompressedCell(PC).SubsiteIndex := Result;
         End;
End;

Function TSite.GetSubSiteNum(X,Y:Integer): Integer;
Var i: Integer;
Begin
  Result := 0;
  For i:=NSubSites downto 1  do
   With SubSites[i-1] do
    If Poly.InPoly(y,x) then
      Begin
        Result := i;
        Exit;
      End;
End;

Function TSite.GetSubSite(X,Y:Integer; PC: pointer): TSubSite;  {Subsite with highest number gets dominance}
Var i: Integer;
Begin
  i := GetSubSiteNum(X,Y,PC);
  IF i=0 then Result := GlobalSite
         else Result := SubSites[i-1];
End;

Function TSite.GetSubSite(X,Y:Integer): TSubSite;  {Subsite with highest number gets dominance}
Var i: Integer;
Begin
  i := GetSubSiteNum(X,Y);
  IF i=0 then Result := GlobalSite
         else Result := SubSites[i-1];

End;



Function TSite.T0: Integer; {Latest NWI_Photo_Date}
Var i, Latest: Integer;
Begin
  Latest := GlobalSite.NWI_Photo_Date;
  For i:=0 to NSubSites-1 do
   If SubSites[i].NWI_Photo_Date > Latest then
     Latest := SubSites[i].NWI_Photo_Date;

  Result := Latest;
End;

Function TSite.TideOc(X,Y: Integer): Double;
Begin
  TideOc := GetSubSite(X,Y).GTideRange;
End;


{**************************************************}
{    Set site constants for elev. ranges           }
{         R. A. Park    3/16/88                    }
{ MOD JSC, Change Ranges based on LIDAR, 2006-2007 }
{**************************************************}
Procedure TSite.InitElevVars;

  Procedure SetSubSiteRanges(Var SS: TSubSite);
  Begin
   With SS do
     Begin
       MLLW := MTL - GTideRange/2.0;    {Mean Lower Low Water}
       MHHW := GTideRange/2.0;          {Mean Higher High Water}
     End;
  End;

Var i: Integer;
Begin
  SetSubSiteRanges(GlobalSite);
  For i := 0 to NSubSites-1 do
    SetSubSiteRanges(SubSites[i]);

End;


Procedure TSite.AddOutputSite;
Begin
  Inc(NOutputSites);
  If NOutputsites > Length(OutputSites) then
    SetLength(OutputSites,NOutputsites+5);

  With OutputSites[NOutputsites-1] do
    Begin
      Description := 'OutputSite '+IntToStr(NOutputSites);
      Poly := nil;
      UsePolygon := False; //fixme allow for user drawn polygon outputs sites eventually
      Rec.X1 := -99;
      Rec.Y1 := -99;
      Rec.X2 := -99;
      Rec.Y2 := -99;
    End;

End;


Procedure TSite.AddSubSite;
Begin
  Inc(NSubSites);
  If NSubsites > Length(SubSites) then
    SetLength(SubSites,NSubsites+5);

  SubSites[NSubsites-1] := TSubSite.Createwith(GlobalSite);
  With SubSites[NSubsites-1] do
    Begin
      Description := 'SubSite '+IntToStr(NSubSites);
      Poly.NumPts := 0;
      Poly.TPoints := nil;
    End;

End;

Procedure TSite.DelSubSite(Index: Integer); {indexed to zero}
Var i: Integer;
Begin
  SubSites[Index].Description := 'Baleeted';
  SubSites[Index].Destroy;
  For i:= Index to NSubSites-2 do
     SubSites[i] := SubSites[i+1];
  Dec(NSubSites);
End;

Procedure TSite.DelOutputSite(Index: Integer); {indexed to zero}
Var i: Integer;
Begin
  OutputSites[NOutputSites-1].Description := 'Baleeted';
  For i:= Index to NOutputSites-2 do
     OutputSites[i] := OutputSites[i+1];
  Dec(NOutputSites);
End;


Function CopyFile(Src,Dst: String): Boolean;
{Copy a file from the source to the destination paths indicated,
 Copied out of Delphi Help under BlockRead}

        procedure CardinalsToI64(var I: Int64; const LowPart, HighPart: Cardinal);
        begin
          TULargeInteger(I).LowPart := LowPart;
          TULargeInteger(I).HighPart := HighPart;
        end;

Var Source,Dest: File;
    NumRead,NumWritten: Integer;
    TotalNumWritten: Int64;
    Buf: array[1..32768] of Char;

Begin
  CopyFile:=True;
  TotalNumWritten:=0;
     Try
         AssignFile(Source,Src);
         AssignFile(Dest,Dst);
         Reset(Source,1);

         Rewrite(Dest,1);
         Repeat
           Blockread(Source,Buf,Sizeof(Buf),NumRead);
           Blockwrite(Dest,Buf,NumRead,NumWritten);
           TotalNumWritten:=TotalNumWritten+NumWritten;
         Until (NumRead=0) or (NumWritten<>NumRead);
         System.CloseFile(Source);
         System.CloseFile(Dest);
      Except
         Raise ESLAMMERROR.Create('File Copy Error:  '+src+'  to  '+dst);
         CopyFile:=False;
      end; {try}
End;

{ TSubSite }

constructor TSubSite.Create;
Var i: Integer;
begin
   Poly := TPolygon.Create;

  OW_Max_Width_Overwash := 500; {m}
  OW_Beach_to_Ocean   := 30;
  OW_Dryland_to_Beach := 30;
  OW_Est_to_Beach     := 60;
  OW_Marsh_Pct_Loss   := 50;
  OW_Mang_Pct_Loss    := 25;

  For i := 0 to N_ACCR_MODELS-1 do
     Begin
       UseAccrModel[i] := False;  {boolean}
       MaxAccr[i] := 0;  {mm/year}
       MinAccr[i] := 0; {mm/year}
       AccrA[i] := 0; {unitless}
       AccrB[i] := 0; {unitless}
       AccrC[i] := 0; {unitless}
       DistEffectMax[i] := 0; {meters}
       Dmin[i] := 1; {unitless scaling factor, set to 1.0 to ignore distance factor}
       SalinityTMax[i] := 0; {ppt}
       STMaxZone[i] := 0; {ppt}
       SNonTMax[i] := 1; {unitless multiplier, set to one to ignore salinity}
       AccrNotes[i] := '';
     End;
end;

constructor TSubSite.CreateWith(TSS: TSubSite);
begin
    Poly := TPolygon.Create;
    Description    := TSS.Description;
    NWI_Photo_Date := TSS.NWI_Photo_Date;
    DEM_date       := TSS.DEM_date;
    Direction_OffShore   := TSS.Direction_OffShore;
    Historic_trend       := TSS.Historic_trend;
    NAVD88MTL_correction := TSS.NAVD88MTL_correction;
    GTideRange  := TSS.GTideRange;
    SaltElev    := TSS.SaltElev;
    MarshErosion := TSS.MarshErosion;
    SwampErosion := TSS.SwampErosion;
    TFlatErosion   := TSS.TFlatErosion;
    FixedRegFloodAccr := TSS.FixedRegFloodAccr;
    FixedIrregFloodAccr := TSS.FixedIrregFloodAccr;
    FixedTideFreshAccr := TSS.FixedTideFreshAccr;

    InlandFreshAccr:= TSS.InlandFreshAccr;
    MangroveAccr:= TSS.MangroveAccr;
    TSwampAccr:= TSS.TSwampAccr;
    SwampAccr:= TSS.SwampAccr;

    Fixed_TF_Beach_Sed := TSS.Fixed_TF_Beach_Sed;
    Freq_BigStorm := TSS.Freq_BigStorm;
    Use_Preprocessor := TSS.Use_Preprocessor;
    OW_Max_Width_Overwash := TSS.OW_Max_Width_Overwash;
    OW_Beach_to_Ocean := TSS.OW_Beach_to_Ocean;
    OW_Dryland_to_Beach := TSS.OW_Dryland_to_Beach;
    OW_Est_to_Beach := TSS.OW_Est_to_Beach;
    OW_Marsh_Pct_Loss := TSS.OW_Marsh_Pct_Loss;
    OW_Mang_Pct_Loss := TSS.OW_Mang_Pct_Loss;
    UseAccrModel := TSS.UseAccrModel;
    MaxAccr := TSS.MaxAccr;
    MinAccr := TSS.MinAccr;
    AccrA := TSS.AccrA;
    AccrB := TSS.AccrB;
    AccrC := TSS.AccrC;
    DistEffectMax := TSS.DistEffectMax;
    Dmin := TSS.Dmin;
    SalinityTMax := TSS.SalinityTMax;
    STMaxZone := TSS.STMaxZone;
    SNonTMax := TSS.SNonTMax;
    AccrNotes := TSS.AccrNotes;
end;

destructor TSubSite.Destroy;
begin
   If Poly<>nil then Poly.Free;
   Poly := nil;
end;


Constructor TSalinityRule.Create;
Begin
  FromCat := Blank;
  ToCat := Blank;
  SalinityLevel := 30;
  GreaterThan := True;
  SalinityTide := 3;
  Descript := '';
End;


Constructor TSalinityRule.Load(ReadVersionNum: Double; TS: TStream);
Var FromInt, ToInt: Integer;
Begin
  TSRead('FromInt',FromInt);
  FromCat := Category(FromInt);
  TSRead('ToInt',ToInt);
  ToCat := Category(ToInt);
  TSRead('SalinityLevel',SalinityLevel);
  If ReadVersionNum > 6.075
    then TSRead('GreaterThan',GreaterThan)
    else GreaterThan := True;
  TSRead('SalinityTide',SalinityTide);
  TSRead('Descript',Descript,ReadVersionNum);
End;

Procedure TSalinityRule.Store(Var TS: TStream);
Var FromInt, ToInt: Integer;
Begin
  FromInt := ORD(FromCat);
  TSWrite('FromInt',FromInt);
  ToInt := ORD(ToCat);
  TSWrite('ToInt',ToInt);
  ToCat := Category(ToInt);
  TSWrite('SalinityLevel',SalinityLevel);
  TSWrite('GreaterThan',GreaterThan);
  TSWrite('SalinityTide',SalinityTide);
  TSWrite('Descript',Descript);

End;

Function TSalinityRules.FindRule(Cat: Category; RuleNum:Integer): TSalinityRule;
Var i,j: Integer;
begin
  Result := nil;
  j := 0;
  for i := 0 to NRules-1 do
    if Rules[i].FromCat = Cat then
      begin
        Inc(j);
        if j=RuleNum then
          Begin
            Result := Rules[i];
            Exit;
          End;
      end;
end;

Function TSalinityRules.CountRules(Cat: Category): Integer;
Var i: Integer;
begin
  Result := 0;
  for i := 0 to NRules-1 do
    if Rules[i].FromCat = Cat then inc(Result);
End;


Constructor TSalinityRules.Create;
Begin
   NRules := 0;
   Rules := nil;

End;

Constructor TSalinityRules.Load(ReadVersionNum: Double; TS: TStream);
Var i: Integer;
Begin
  TSRead('NRules',NRules);
  SetLength(Rules,NRules);
  For i := 0 to NRules-1 do
    Rules[i] := TSalinityRule.Load(ReadVersionNum, TS);

End;

Procedure TSalinityRules.Store(Var TS: TStream);
Var i: Integer;
Begin
  TSWrite('NRules',NRules);
  For i := 0 to NRules-1 do
    Rules[i].Store(TS);
End;

{-----------------------------------------------------------------------------}
Procedure LoadElevRangesFromText(Var TCA: TClassElevArray);
Var Cat: Category;
   UnitStr: String;

    Function Str2Unit(UStr: String): ElevUnit;
    Begin
      UStr := Lowercase(UStr);
      If Pos(UStr,'halftide')>0 then Result := HalfTide
         else If Pos(UStr,'saltbound')>0 then Result := SaltBound
         else Result := Meters;
    End;                                                                                

Begin
  For Cat := FirstCategory to LastCategory do
    Begin
      TSRead(TitleCat[Cat]+' MinElev',TCA[Cat].MinElev);
      TSRead(TitleCat[Cat]+' MinUnit',UnitStr,999);
      TCA[Cat].MinUnit := Str2Unit(UnitStr);

      TSRead(TitleCat[Cat]+' MaxElev',TCA[Cat].MaxElev);
      TSRead(TitleCat[Cat]+' MaxUnit',UnitStr,999);
      TCA[Cat].MaxUnit := Str2Unit(UnitStr);
    End;
End;
{-----------------------------------------------------------------------------}

Procedure SaveElevRangesToText(TCA: TClassElevArray);
Var Cat: Category;
   UnitStr: String;

    Function Unit2Str(EU:ElevUnit): String;
    Begin
      Case EU of
        HalfTide: Result :='HalfTide';
        SaltBound: Result := 'SaltBound';
        else Result := 'Meters';
      end; {case}
    End;

Begin
  For Cat := FirstCategory to LastCategory do
    Begin
      TSWrite(TitleCat[Cat]+' MinElev',TCA[Cat].MinElev);
      UnitStr := Unit2Str(TCA[Cat].MinUnit);
      TSWrite(TitleCat[Cat]+' MinUnit',UnitStr);

      TSWrite(TitleCat[Cat]+' MaxElev',TCA[Cat].MaxElev);
      UnitStr := Unit2Str(TCA[Cat].MaxUnit);
      TSWrite(TitleCat[Cat]+' MaxUnit',UnitStr);
    End;
End;

{-----------------------------------------------------------------------------}

Function IntInString(I: Integer; Str: String): Boolean;
Var NumberFound, EofFound: Boolean;
    Holder: String;
    Number: Integer;
Begin
  EofFound := False;
  NumberFound := False;
   Repeat
    Holder:=AbbrString(Str,',');
    Number := StrToInt(Holder);
    If Number = I then NumberFound:=True;
    If Pos(',',Str)= 0 then EofFound:=True
                             else Delete(Str,1,Pos(',',Str));
    If Str='' then EofFound := True;
   Until EofFound or Numberfound;

   Result:=NumberFound;
End;


end.
